/**
   @author Shin'ichiro Nakaoka
*/

#include "ValueTree.h"
#include <stack>
#include <iostream>
#include <yaml.h>
#include <boost/lexical_cast.hpp>
#include <boost/format.hpp>
#include <boost/filesystem.hpp>

#ifdef _WIN32
#define snprintf _snprintf_s
#endif

using namespace std;
using namespace boost;
using namespace cnoid;


namespace {
    const bool debugTrace = false;
    const char* typeNames[] = { "unknown node", "mapping", "sequence", "scalar" };
    map<string, bool> booleanSymbols;

    ValueNodePtr invalidNode;
    MappingPtr invalidMapping;
    SequencePtr invalidSequence;
}

ValueNode::Initializer ValueNode::initializer;

ValueNode::Initializer::Initializer()
{
    invalidNode = new ValueNode(INVALID_NODE);
    invalidMapping = new Mapping();
    invalidMapping->type_ = INVALID_NODE;
    invalidSequence = new Sequence();
    invalidSequence->type_ = INVALID_NODE;
    
    booleanSymbols["true"] = true;
    booleanSymbols["yes"] = true;
    booleanSymbols["on"] = true;
    booleanSymbols["false"] = false;
    booleanSymbols["no"] = false;
    booleanSymbols["off"] = false;
}


ValueNode::Exception::~Exception()
{

}


// disabled
ValueNode::ValueNode(const ValueNode&)
{
    // throw an exception here ?
}


// disabled
ValueNode& ValueNode::operator=(const ValueNode&)
{
    // throw an exception here ?
    return *this;
}

/*
void ValueNode::initialize()
{
    static bool initialized = false;

    if(!initialized){

        invalidNode = new ValueNode(INVALID_NODE);

        invalidMapping = new Mapping();
        invalidMapping->type_ = INVALID_NODE;

        invalidSequence = new Sequence();
        invalidSequence->type_ = INVALID_NODE;
        
        booleanSymbols["true"] = true;
        booleanSymbols["yes"] = true;
        booleanSymbols["on"] = true;
        booleanSymbols["false"] = false;
        booleanSymbols["no"] = false;
        booleanSymbols["off"] = false;

        initialized = true;
    }
}
*/


bool ValueNode::read(int &out_value) const
{
    if(type_ == SCALAR){
        const char* nptr = &(static_cast<const ScalarNode* const>(this)->stringValue[0]);
        char* endptr;
        out_value = strtol(nptr, &endptr, 10);
        if(endptr > nptr){
            return true;
        }
    }
    return false;
}


int ValueNode::toInt() const
{
    if(type_ != SCALAR){
        throwNotScalrException();
    }

    const ScalarNode* const scalar = static_cast<const ScalarNode* const>(this);
    
    const char* nptr = &(scalar->stringValue[0]);
    char* endptr;
    const int value = strtol(nptr, &endptr, 10);

    if(endptr == nptr){
        ScalarTypeMismatchException ex;
        ex.setMessage(str(format("\"%1%\" at line %2%, column %3% should be an integer value.")
                      % scalar->stringValue % line() % column()));
        ex.setPosition(line(), column());
        throw ex;
    }

    return value;
}


double ValueNode::toDouble() const
{
    if(type_ != SCALAR){
        throwNotScalrException();
    }

    const ScalarNode* const scalar = static_cast<const ScalarNode* const>(this);

    const char* nptr = &(scalar->stringValue[0]);
    char* endptr;
    const double value = strtod(nptr, &endptr);

    if(endptr == nptr){
        ScalarTypeMismatchException ex;
        ex.setMessage(str(format("\"%1%\" at line %2%, column %3% should be a double value.")
                          % scalar->stringValue % line() % column()));
        ex.setPosition(line(), column());
        throw ex;
    }

    return value;
}


bool ValueNode::toBool() const
{
    if(type_ != SCALAR){
        throwNotScalrException();
    }

    const ScalarNode* const scalar = static_cast<const ScalarNode* const>(this);

    map<string, bool>::iterator p = booleanSymbols.find(scalar->stringValue);
    if(p != booleanSymbols.end()){
        return p->second;
    }
    
    ScalarTypeMismatchException ex;
    ex.setMessage(str(format("\"%1%\" at line %2%, column %3% should be a bool value.")
                      % scalar->stringValue % line() % column()));
    ex.setPosition(line(), column());
    throw ex;
}


#ifdef _WIN32
const std::string ValueNode::toString() const
{
    if(type_ != SCALAR){
        throwNotScalrException();
    }
    return fromUTF8(static_cast<const ScalarNode* const>(this)->stringValue);
}

const std::string ValueNode::toUTF8String() const
{
    if(type_ != SCALAR){
        throwNotScalrException();
    }
    return static_cast<const ScalarNode* const>(this)->stringValue;
}    

#else

const std::string& ValueNode::toString() const
{
    if(type_ != SCALAR){
        throwNotScalrException();
    }
    return static_cast<const ScalarNode* const>(this)->stringValue;
}


const std::string& ValueNode::toUTF8String() const
{
    return toString();
}
#endif


ScalarNode::ScalarNode(const std::string& value, StringStyle stringStyle)
    : stringValue(value),
      stringStyle(stringStyle)
{
    type_ = SCALAR;
    line_ = -1;
    column_ = -1;
}


ScalarNode::ScalarNode(const char* text, size_t length)
    : stringValue(text, length)
{
    type_ = SCALAR;
    stringStyle = PLAIN_STRING;
}


ScalarNode::ScalarNode(const char* text, size_t length, StringStyle stringStyle)
    : stringValue(text, length),
      stringStyle(stringStyle)
{
    type_ = SCALAR;
    line_ = -1;
    column_ = -1;
}


const Mapping* ValueNode::toMapping() const
{
    if(type_ != MAPPING){
        throwNotMappingException();
    }
    return static_cast<const Mapping*>(this);
}


Mapping* ValueNode::toMapping()
{
    if(type_ != MAPPING){
        throwNotMappingException();
    }
    return static_cast<Mapping*>(this);
}


const Sequence* ValueNode::toSequence() const
{
    if(type_ != SEQUENCE){
        throwNotSequenceException();
    }
    return static_cast<const Sequence*>(this);
}


Sequence* ValueNode::toSequence()
{
    if(type_ != SEQUENCE){
        throwNotSequenceException();
    }
    return static_cast<Sequence*>(this);
}


void ValueNode::throwException(const std::string& message) const
{
    Exception ex;
    
    if(hasLineInfo()){
        ex.setMessage(str(format("%1% (line %2%, column %3%).")
                      % message % line() % column()));
    } else {
        ex.setMessage(message);
    }
    ex.setPosition(line(), column());
    throw ex;
}


void ValueNode::throwNotScalrException() const
{
    NotScalarException ex;
    if(hasLineInfo()){
        ex.setMessage(str(format("The %1% at line %2%, column %3% should be a scalar value.")
                      % typeNames[type_] % line() % column()));
    } else {
        ex.setMessage("Scalar value cannot be obtained from a non-scalar type yaml node.");
    }
    ex.setPosition(line(), column());
    throw ex;
}


void ValueNode::throwNotMappingException() const
{
    NotMappingException ex;
    ex.setPosition(line(), column());
    throw ex;
}


void ValueNode::throwNotSequenceException() const
{
    NotSequenceException ex;
    ex.setPosition(line(), column());
    throw ex;
}


Collection::Collection()
{
    isFlowStyle_ = false;
    static const char* defaultFormat = "%.6g";
    doubleFormat_ = defaultFormat;
}

    
Collection::~Collection()
{

}


void Collection::setDoubleFormat(const char* format)
{
    doubleFormat_ = format;
}


Mapping::Mapping()
{
    type_ = MAPPING;
    line_ = -1;
    column_ = -1;
    mode = READ_MODE;
    indexCounter = 0;
    keyQuoteStyle = PLAIN_STRING;
}


Mapping::Mapping(int line, int column)
{
    type_ = MAPPING;
    line_ = line;
    column_ = column;
    mode = READ_MODE;
    indexCounter = 0;
}


Mapping::~Mapping()
{
    clear();
}


void Mapping::clear()
{
    values.clear();
    indexCounter = 0;
}


void Mapping::setKeyQuoteStyle(StringStyle style)
{
    keyQuoteStyle = style;
}


ValueNode* Mapping::find(const std::string& key) const
{
    if(!isValid()){
        throwNotMappingException();
    }
    const_iterator p = values.find(toUTF8(key));
    if(p != values.end()){
        return p->second.get();
    } else {
        return invalidNode.get();
    }
}


Mapping* Mapping::findMapping(const std::string& key) const
{
    if(!isValid()){
        throwNotMappingException();
    }
    const_iterator p = values.find(toUTF8(key));
    if(p != values.end()){
        ValueNode* node = p->second.get();
        if(node->type() == MAPPING){
            return static_cast<Mapping*>(node);
        }
    }
    return invalidMapping.get();
}


Sequence* Mapping::findSequence(const std::string& key) const
{
    if(!isValid()){
        throwNotMappingException();
    }
    const_iterator p = values.find(toUTF8(key));
    if(p != values.end()){
        ValueNode* node = p->second.get();
        if(node->type() == SEQUENCE){
            return static_cast<Sequence*>(node);
        }
    }
    return invalidSequence.get();
}


ValueNode& Mapping::get(const std::string& key) const
{
    if(!isValid()){
        throwNotMappingException();
    }
    const_iterator p = values.find(toUTF8(key));
    if(p == values.end()){
        throwKeyNotFoundException(key);
    }
    return *p->second;
}


void Mapping::throwKeyNotFoundException(const std::string& key) const
{
    KeyNotFoundException ex;
    ex.setMessage(str(format("Key \"%1%\" is not found in the mapping that begins at line %2%, column %3%.")
                      % key % line() % column()));
    ex.setPosition(line(), column());
    ex.setKey(key);
    throw ex;
}


inline void Mapping::insertSub(const std::string& key, ValueNode* node)
{
    values.insert(make_pair(key, node));
    node->indexInMapping = indexCounter++;
}


void Mapping::insert(const std::string& key, ValueNodePtr node)
{
    if(!isValid()){
        throwNotMappingException();
    }
    const string uKey(toUTF8(key));
    iterator p = values.find(uKey);
    if(p != values.end()){
        values.erase(p);
    }
    insertSub(uKey, node.get());
}


Mapping* Mapping::openMapping(const std::string& key, bool doOverwrite)
{
    if(!isValid()){
        throwNotMappingException();
    }

    Mapping* mapping = 0;
    const string uKey(toUTF8(key));
    iterator p = values.find(uKey);
    if(p != values.end()){
        ValueNode* node = p->second.get();
        if(node->type() != MAPPING){
            values.erase(p);
        } else {
            mapping = static_cast<Mapping*>(node);
            if(doOverwrite){
                mapping->clear();
            }
            mapping->indexInMapping = indexCounter++;
        }
    }

    if(!mapping){
        mapping = new Mapping();
        mapping->doubleFormat_ = doubleFormat_;
        insertSub(uKey, mapping);
    }

    return mapping;
}


Mapping* Mapping::openFlowStyleMapping(const std::string& key, bool doOverwrite)
{
    Mapping* m = openMapping(key, doOverwrite);
    m->setFlowStyle(true);
    return m;
}


Sequence* Mapping::openSequence(const std::string& key, bool doOverwrite)
{
    if(!isValid()){
        throwNotMappingException();
    }

    Sequence* sequence = 0;
    const string uKey(toUTF8(key));
    iterator p = values.find(uKey);
    if(p != values.end()){
        ValueNode* node = p->second.get();
        if(node->type() != SEQUENCE){
            values.erase(p);
        } else {
            sequence = static_cast<Sequence*>(node);
            if(doOverwrite){
                sequence->clear();
            }
            sequence->indexInMapping = indexCounter++;
        }
    }

    if(!sequence){
        sequence = new Sequence();
        sequence->doubleFormat_ = doubleFormat_;
        insertSub(uKey, sequence);
    }

    return sequence;
}


Sequence* Mapping::openFlowStyleSequence(const std::string& key, bool doOverwrite)
{
    Sequence* s = openSequence(key, doOverwrite);
    s->setFlowStyle(true);
    return s;
}


bool Mapping::read(const std::string &key, std::string &out_value) const
{
    ValueNode* node = find(toUTF8(key));
    if(node->isValid()){
        out_value = node->toString();
        return true;
    }
    return false;
}


bool Mapping::readUTF8(const std::string &key, std::string &out_value) const
{
    ValueNode* node = find(toUTF8(key));
    if(node->isValid()){
        out_value = node->toUTF8String();
        return true;
    }
    return false;
}


bool Mapping::read(const std::string &key, bool &out_value) const
{
    ValueNode* node = find(toUTF8(key));
    if(node->isValid()){
        out_value = node->toBool();
        return true;
    }
    return false;
}


bool Mapping::read(const std::string &key, int &out_value) const
{
    ValueNode* node = find(toUTF8(key));
    if(node->isValid()){
        out_value = node->toInt();
        return true;
    }
    return false;
}


bool Mapping::read(const std::string &key, double &out_value) const
{
    ValueNode* node = find(toUTF8(key));
    if(node->isValid()){
        out_value = node->toDouble();
        return true;
    }
    return false;
}


bool Mapping::read(const std::string &key, boost::function<void(double)> setterFunc) const
{
    double value;
    if(read(key, value)){
        setterFunc(value);
        return true;
    }
    return false;
}


void Mapping::writeUTF8(const std::string &key, const std::string& value, StringStyle stringStyle)
{
    string uKey(toUTF8(key));
    iterator p = values.find(uKey);
    if(p == values.end()){
        insertSub(uKey, new ScalarNode(value, stringStyle));
    } else {
        ValueNode* node = p->second.get();
        if(node->type() == SCALAR){
            ScalarNode* scalar = static_cast<ScalarNode*>(node);
            scalar->stringValue = value;
            scalar->stringStyle = stringStyle;
            scalar->indexInMapping = indexCounter++;
        } else {
            throwNotScalrException();
        }
    }
}


/**
   This is for internal use. Text are not converted to UTF-8.
*/
void Mapping::writeSub(const std::string &key, const char* text, size_t length, StringStyle stringStyle)
{
    const string uKey(toUTF8(key));
    iterator p = values.find(uKey);
    if(p == values.end()){
        insertSub(uKey, new ScalarNode(text, length, stringStyle));
    } else {
        ValueNode* node = p->second.get();
        if(node->type() == SCALAR){
            ScalarNode* scalar = static_cast<ScalarNode*>(node);
            scalar->stringValue = string(text, length);
            scalar->stringStyle = stringStyle;
            scalar->indexInMapping = indexCounter++;
        } else {
            throwNotScalrException();
        }
    }
}


void Mapping::write(const std::string &key, bool value)
{
    if(value){
        writeSub(key, "true", 4, PLAIN_STRING);
    } else {
        writeSub(key, "false", 5, PLAIN_STRING);
    }
}


void Mapping::write(const std::string &key, int value)
{
    char buf[32];
    int n = snprintf(buf, 32, "%d", value);
    writeSub(key, buf, n, PLAIN_STRING);
}


void Mapping::write(const std::string &key, double value)
{
    char buf[32];
    int n = snprintf(buf, 32, doubleFormat_, value);
    writeSub(key, buf, n, PLAIN_STRING);
}


void Mapping::writePath(const std::string &key, const std::string& value)
{
    write(key, filesystem::path(value).string(), DOUBLE_QUOTED);
}


bool Mapping::compareIters(const Mapping::const_iterator& it1, const Mapping::const_iterator& it2)
{
    return (it1->second->indexInMapping < it2->second->indexInMapping);
}


Sequence::Sequence()
{
    type_ = SEQUENCE;
    line_ = -1;
    column_ = -1;
}


Sequence::Sequence(int size)
    : values(size)
{
    type_ = SEQUENCE;
    line_ = -1;
    column_ = -1;
}


Sequence::Sequence(int line, int column)
{
    type_ = SEQUENCE;
    line_ = line;
    column_ = column;
}


Sequence::Sequence(int line, int column, int reservedSize)
    : values(reservedSize)
{
    type_ = SEQUENCE;
    line_ = line;
    column_ = column;
    values.resize(0);
}


Sequence::~Sequence()
{
    clear();
}


void Sequence::clear()
{
    values.clear();
}


void Sequence::reserve(int size)
{
    values.reserve(size);
}


void Sequence::insertLF(int maxColumns, int numValues)
{
    if(values.empty()){
        if(numValues > 0 && numValues > maxColumns){
            appendLF();
        }
    } else if((values.size() % maxColumns) == 0){
        appendLF();
    }
}


Mapping* Sequence::newMapping()
{
    Mapping* mapping = new Mapping();
    mapping->doubleFormat_ = doubleFormat_;
    append(mapping);
    return mapping;
}


void Sequence::append(int value)
{
    char buf[32];
    int n = snprintf(buf, 32, "%d", value);
    values.push_back(new ScalarNode(buf, n, PLAIN_STRING));
}


void Sequence::write(int i, int value)
{
    char buf[32];
    int n = snprintf(buf, 32, "%d", value);
    values[i] = new ScalarNode(buf, n, PLAIN_STRING);
}


/*
void Sequence::append(size_t value)
{
    char buf[32];
    int n = snprintf(buf, 32, "%zd", value);
    values.push_back(new ScalarNode(buf, n, PLAIN_STRING));
}
*/


void Sequence::append(double value)
{
    char buf[32];
    int n = snprintf(buf, 32, doubleFormat_, value);
    values.push_back(new ScalarNode(buf, n, PLAIN_STRING));
}


void Sequence::append(const std::string& value, StringStyle stringStyle)
{
    values.push_back(new ScalarNode(toUTF8(value), stringStyle));
}


void Sequence::write(int i, const std::string& value, StringStyle stringStyle)
{
    values[i] = new ScalarNode(toUTF8(value), stringStyle);
}


void Sequence::appendLF()
{
    values.push_back(new ValueNode(LF_NODE));
}
